class ParclLabsException(Exception):
    """Base exception class for ParclLabs SDK"""

    pass


class NotFoundError(ParclLabsException):
    """Exception raised when no data is found matching search criteria (404 error)."""

    def __init__(
        self,
        message="No data found matching search criteria. Try a different set of parameters.",
        *args,
        **kwargs,
    ):
        super().__init__(message, *args, **kwargs)


class DataValidationError(ParclLabsException):
    """Exception raised for data validation errors (422 error)."""

    def __init__(
        self, message="Data validation error occurred.", details=None, *args, **kwargs
    ):
        self.details = details
        super().__init__(message, *args, **kwargs)

    def __str__(self):
        if self.details:
            return f"{super().__str__()}\nDetails: {self.details}"
        return super().__str__()



import pandas as pd
import numpy as np

from parcllabs.common import ID_COLUMNS, DATE_COLUMNS


def safe_concat_and_format_dtypes(data_container):
    # Filter out empty DataFrames
    non_empty_dfs = [df for df in data_container if not df.empty]

    # Check if there are any DataFrames left to concatenate
    if not non_empty_dfs:
        return (
            pd.DataFrame()
        )  # Return an empty DataFrame if no non-empty DataFrames are found

    # Get the column order from the first non-empty DataFrame
    original_columns = non_empty_dfs[0].columns.tolist()

    # Find common columns across all non-empty DataFrames
    common_columns = list(set.intersection(*[set(df.columns) for df in non_empty_dfs]))

    # Initialize an empty list to store processed DataFrames
    processed_dfs = []

    for df in non_empty_dfs:
        # Select only common columns
        df = df[common_columns].copy()

        # Remove columns that are entirely NA
        df = df.dropna(axis=1, how="all")

        # For columns that are all empty strings, replace with NaN
        for col in df.columns:
            if df[col].dtype == object and df[col].astype(str).str.strip().eq("").all():
                df[col] = np.nan

        # Remove columns that became all NaN after replacing empty strings
        df = df.dropna(axis=1, how="all")

        if not df.empty:
            processed_dfs.append(df)

    # Concatenate the processed DataFrames
    if processed_dfs:
        output = pd.concat(processed_dfs, axis=0, ignore_index=True)
    else:
        return (
            pd.DataFrame()
        )  # Return an empty DataFrame if all processed DataFrames are empty

    # Cast date columns to datetime
    for col in DATE_COLUMNS:
        if col in output.columns:
            output[col] = pd.to_datetime(output[col], errors="coerce")

    # Reorder columns based on the specified rules and original order
    final_columns = []

    # Rule 1: parcl_id or parcl_property_id should be the first column
    for col in ID_COLUMNS:
        if col in output.columns:
            final_columns.append(col)
            break

    # Rule 2: date or event_date should be the second column
    date_columns = ["date", "event_date"]
    for col in date_columns:
        if col in output.columns:
            final_columns.append(col)
            break

    # Add remaining columns in their original order
    for col in original_columns:
        if col in output.columns and col not in final_columns:
            final_columns.append(col)

    # Reorder the DataFrame
    output = output[final_columns]

    return output



from typing import Any, Mapping, Optional

from parcllabs.services.parcllabs_service import ParclLabsService
from parcllabs.services.validators import Validators


class PropertyTypeService(ParclLabsService):
    def retrieve(
        self,
        parcl_ids: int,
        start_date: str = None,
        end_date: str = None,
        property_type: str = None,
        limit: Optional[int] = None,
        params: Optional[Mapping[str, Any]] = {},
        auto_paginate: bool = False,
    ):
        property_type = Validators.validate_property_type(property_type)

        if property_type:
            params["property_type"] = property_type

        return super().retrieve(
            parcl_ids=parcl_ids,
            start_date=start_date,
            end_date=end_date,
            limit=limit,
            params=params,
            auto_paginate=auto_paginate,
        )



from parcllabs.enums import *


VALID_PROPERTY_TYPES = get_enum_values(PropertyTypes)
VALID_PROPERTY_TYPES_UNIT_SEARCH = get_enum_values(PropertyTypesUnit)
VALID_ENTITY_NAMES = get_enum_values(EntityNames)
VALID_PORTFOLIO_SIZES = get_enum_values(PortfolioSizes)
VALID_LOCATION_TYPES = get_enum_values(LocationTypes)
VALID_US_REGIONS = get_enum_values(USRegions)
VALID_US_STATE_ABBREV = get_enum_values(USStateAbbreviations)
VALID_US_STATE_FIPS_CODES = get_enum_values(USStateFIPSCodes)
VALID_SORT_BY = get_enum_values(SortByParams)
VALID_SORT_ORDER = get_enum_values(SortOrder)
VALID_EVENT_TYPES = get_enum_values(EventTypes)

ID_COLUMNS = [ResponseColumns.PARCL_ID.value, ResponseColumns.PARCL_PROPERTY_ID.value]

DATE_COLUMNS = [ResponseColumns.DATE.value, ResponseColumns.EVENT_DATE.value]

DELETE_FROM_OUTPUT = [
    ResponseColumns.TOTAL.value,
    ResponseColumns.LIMIT.value,
    ResponseColumns.OFFSET.value,
    ResponseColumns.LINKS.value,
    ResponseColumns.ACCOUNT.value,
]

PARCL_LABS_DASHBOARD_URL = "https://dashboard.parcllabs.com/signup"
NO_API_KEY_ERROR = (
    f"API Key is required. Please visit {PARCL_LABS_DASHBOARD_URL} to get an API key."
)

GET_METHOD = RequestMethods.GET.value
POST_METHOD = RequestMethods.POST.value



from typing import List, Dict

import pandas as pd

from parcllabs.common import VALID_US_STATE_ABBREV
from parcllabs.services.validators import Validators
from parcllabs.services.parcllabs_service import ParclLabsService


class PropertyAddressSearch(ParclLabsService):
    """
    Retrieve parcl_property_ids based on provided addresses.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retrieve(
        self,
        addresses: List[Dict],
    ):
        """
        Retrieve parcl_property_ids based on provided addresses.

        Args:

            addresses (list): A list of dictionaries containing address information.

        Returns:

            DataFrame: A DataFrame containing the parcl_property_id and address information.
        """

        required_params = ["address", "city", "state_abbreviation", "zip_code"]
        for address in addresses:
            param = Validators.validate_field_exists(address, required_params)

            param = Validators.validate_input_str_param(
                param=address.get("state_abbreviation"),
                param_name="state_abbreviation",
                valid_values=VALID_US_STATE_ABBREV,
                params_dict={},
            )
            param = Validators.validate_us_zip_code(zip_code=address.get("zip_code"))

        response = self._post(url=self.full_url, data=addresses)
        results = response.json()
        data = pd.DataFrame(results.get("items"))
        self._update_account_info(results.get("account"))
        return data



from typing import Dict, Any, Optional

from parcllabs import api_base
from parcllabs.common import NO_API_KEY_ERROR
from parcllabs.services.search import SearchMarkets
from parcllabs.services.parcllabs_service import ParclLabsService
from parcllabs.services.metrics.property_type_service import PropertyTypeService
from parcllabs.services.metrics.portfolio_size_service import PortfolioSizeService
from parcllabs.services.properties.property_events_service import PropertyEventsService
from parcllabs.services.properties.property_search import PropertySearch
from parcllabs.services.properties.property_address import PropertyAddressSearch


class ServiceGroup:
    def __init__(self, client):
        self._client = client
        self._services = {}

    def add_service(
        self,
        name: str,
        url: str,
        service_class: Any,
        post_url: Optional[str] = None,
        alias: Optional[str] = None,
    ):
        service = service_class(url=url, post_url=post_url, client=self._client)
        setattr(self, name, service)
        self._services[name] = service
        if alias:
            setattr(self, alias, service)
            self._services[alias] = service

    @property
    def services(self):
        return list(self._services.keys())


class ParclLabsClient:
    def __init__(
        self,
        api_key: str,
        api_url: str = api_base,
        limit: Optional[int] = None,
        turbo_mode: bool = False,
        num_workers: Optional[int] = None,
    ):
        if not api_key:
            raise ValueError(NO_API_KEY_ERROR)

        self.api_key = api_key
        self.api_url = api_url
        self.account_info = {"est_session_credits_used": 0}
        self.num_workers = num_workers
        self.limit = limit
        self.turbo_mode = turbo_mode

        self._initialize_services()

    def _initialize_services(self):
        self.price_feed = self._create_price_feed_services()
        self.investor_metrics = self._create_investor_metrics_services()
        self.market_metrics = self._create_market_metrics_services()
        self.new_construction_metrics = self._create_new_construction_metrics_services()
        self.for_sale_market_metrics = self._create_for_sale_market_metrics_services()
        self.rental_market_metrics = self._create_rental_market_metrics_services()
        self.portfolio_metrics = self._create_portfolio_metrics_services()
        self.search = self._create_search_services()
        self.property = self._create_property_services()
        self.property_address = self._create_property_address_services()

    def _create_service_group(self):
        return ServiceGroup(self)

    def _add_services_to_group(
        self, group: ServiceGroup, services: Dict[str, Dict[str, Any]]
    ):
        for name, config in services.items():
            group.add_service(name=name, **config)

    def _create_price_feed_services(self):
        group = self._create_service_group()
        services = {
            "price_feed": {
                "url": "/v1/price_feed/{parcl_id}/price_feed",
                "post_url": "/v1/price_feed/price_feed",
                "service_class": ParclLabsService,
            },
            "volatility": {
                "url": "/v1/price_feed/{parcl_id}/volatility",
                "post_url": "/v1/price_feed/volatility",
                "service_class": ParclLabsService,
            },
            "rental_price_feed": {
                "url": "/v1/price_feed/{parcl_id}/rental_price_feed",
                "post_url": "/v1/price_feed/rental_price_feed",
                "service_class": ParclLabsService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_investor_metrics_services(self):
        group = self._create_service_group()
        services = {
            "housing_stock_ownership": {
                "url": "/v1/investor_metrics/{parcl_id}/housing_stock_ownership",
                "post_url": "/v1/investor_metrics/housing_stock_ownership",
                "service_class": ParclLabsService,
            },
            "new_listings_for_sale_rolling_counts": {
                "url": "/v1/investor_metrics/{parcl_id}/new_listings_for_sale_rolling_counts",
                "post_url": "/v1/investor_metrics/new_listings_for_sale_rolling_counts",
                "service_class": PropertyTypeService,
            },
            "purchase_to_sale_ratio": {
                "url": "/v1/investor_metrics/{parcl_id}/purchase_to_sale_ratio",
                "post_url": "/v1/investor_metrics/purchase_to_sale_ratio",
                "service_class": ParclLabsService,
            },
            "housing_event_counts": {
                "url": "/v1/investor_metrics/{parcl_id}/housing_event_counts",
                "post_url": "/v1/investor_metrics/housing_event_counts",
                "service_class": ParclLabsService,
            },
            "housing_event_prices": {
                "url": "/v1/investor_metrics/{parcl_id}/housing_event_prices",
                "post_url": "/v1/investor_metrics/housing_event_prices",
                "service_class": ParclLabsService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_market_metrics_services(self):
        group = self._create_service_group()
        services = {
            "housing_event_prices": {
                "url": "/v1/market_metrics/{parcl_id}/housing_event_prices",
                "post_url": "/v1/market_metrics/housing_event_prices",
                "service_class": PropertyTypeService,
            },
            "all_cash": {
                "url": "/v1/market_metrics/{parcl_id}/all_cash",
                "post_url": "/v1/market_metrics/all_cash",
                "service_class": PropertyTypeService,
            },
            "housing_stock": {
                "url": "/v1/market_metrics/{parcl_id}/housing_stock",
                "post_url": "/v1/market_metrics/housing_stock",
                "service_class": ParclLabsService,
            },
            "housing_event_counts": {
                "url": "/v1/market_metrics/{parcl_id}/housing_event_counts",
                "post_url": "/v1/market_metrics/housing_event_counts",
                "service_class": PropertyTypeService,
            },
            "housing_event_property_attributes": {
                "url": "/v1/market_metrics/{parcl_id}/housing_event_property_attributes",
                "post_url": "/v1/market_metrics/housing_event_property_attributes",
                "service_class": PropertyTypeService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_new_construction_metrics_services(self):
        group = self._create_service_group()
        services = {
            "housing_event_prices": {
                "url": "/v1/new_construction_metrics/{parcl_id}/housing_event_prices",
                "post_url": "/v1/new_construction_metrics/housing_event_prices",
                "service_class": PropertyTypeService,
            },
            "housing_event_counts": {
                "url": "/v1/new_construction_metrics/{parcl_id}/housing_event_counts",
                "post_url": "/v1/new_construction_metrics/housing_event_counts",
                "service_class": PropertyTypeService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_for_sale_market_metrics_services(self):
        group = self._create_service_group()
        services = {
            "new_listings_rolling_counts": {
                "url": "/v1/for_sale_market_metrics/{parcl_id}/new_listings_rolling_counts",
                "post_url": "/v1/for_sale_market_metrics/new_listings_rolling_counts",
                "service_class": PropertyTypeService,
            },
            "for_sale_inventory": {
                "url": "/v1/for_sale_market_metrics/{parcl_id}/for_sale_inventory",
                "post_url": "/v1/for_sale_market_metrics/for_sale_inventory",
                "service_class": PropertyTypeService,
            },
            "for_sale_inventory_price_changes": {
                "url": "/v1/for_sale_market_metrics/{parcl_id}/for_sale_inventory_price_changes",
                "post_url": "/v1/for_sale_market_metrics/for_sale_inventory_price_changes",
                "service_class": PropertyTypeService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_rental_market_metrics_services(self):
        group = self._create_service_group()
        services = {
            "rental_units_concentration": {
                "url": "/v1/rental_market_metrics/{parcl_id}/rental_units_concentration",
                "post_url": "/v1/rental_market_metrics/rental_units_concentration",
                "service_class": PropertyTypeService,
            },
            "gross_yield": {
                "url": "/v1/rental_market_metrics/{parcl_id}/gross_yield",
                "post_url": "/v1/rental_market_metrics/gross_yield",
                "service_class": PropertyTypeService,
            },
            "new_listings_for_rent_rolling_counts": {
                "url": "/v1/rental_market_metrics/{parcl_id}/new_listings_for_rent_rolling_counts",
                "post_url": "/v1/rental_market_metrics/new_listings_for_rent_rolling_counts",
                "service_class": PropertyTypeService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_portfolio_metrics_services(self):
        group = self._create_service_group()
        services = {
            "sf_housing_stock_ownership": {
                "url": "/v1/portfolio_metrics/{parcl_id}/sf_housing_stock_ownership",
                "post_url": "/v1/portfolio_metrics/sf_housing_stock_ownership",
                "service_class": ParclLabsService,
            },
            "sf_new_listings_for_sale_rolling_counts": {
                "url": "/v1/portfolio_metrics/{parcl_id}/sf_new_listings_for_sale_rolling_counts",
                "post_url": "/v1/portfolio_metrics/sf_new_listings_for_sale_rolling_counts",
                "service_class": PortfolioSizeService,
            },
            "sf_new_listings_for_rent_rolling_counts": {
                "url": "/v1/portfolio_metrics/{parcl_id}/sf_new_listings_for_rent_rolling_counts",
                "post_url": "/v1/portfolio_metrics/sf_new_listings_for_rent_rolling_counts",
                "service_class": PortfolioSizeService,
            },
            "sf_housing_event_counts": {
                "url": "/v1/portfolio_metrics/{parcl_id}/sf_housing_event_counts",
                "post_url": "/v1/portfolio_metrics/sf_housing_event_counts",
                "service_class": PortfolioSizeService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_search_services(self):
        group = self._create_service_group()
        services = {
            "markets": {
                "url": "/v1/search/markets",
                "service_class": SearchMarkets,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_property_services(self):
        group = self._create_service_group()
        services = {
            "search": {"url": "/v1/property/search", "service_class": PropertySearch},
            "events": {
                "url": "/v1/property/event_history",
                "post_url": "/v1/property/event_history",
                "service_class": PropertyEventsService,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def _create_property_address_services(self):
        group = self._create_service_group()
        services = {
            "search": {
                "url": "/v1/property/search_address",
                "post_url": "/v1/property/search_address",
                "service_class": PropertyAddressSearch,
            },
        }
        self._add_services_to_group(group, services)
        return group

    def account(self):
        return self.account_info



from collections import deque
from concurrent.futures import ThreadPoolExecutor, as_completed
import json

import pandas as pd

from parcllabs.services.parcllabs_service import ParclLabsService


class ParclLabsStreamingService(ParclLabsService):
    def _convert_text_to_json(self, chunk):
        try:
            return json.loads(chunk)
        except json.JSONDecodeError as e:
            print(f"Error decoding JSON: {e}")
            return None

    def _process_streaming_data(self, data, batch_size=10000, num_workers=None):
        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            chunks = deque(data.strip().split("\n"))
            futures = [
                executor.submit(self._convert_text_to_json, chunk)
                for chunk in chunks
                if chunk
            ]

            buffer = deque()
            for future in as_completed(futures):
                result = future.result()
                if result:
                    buffer.append(result)

                if len(buffer) >= batch_size:
                    yield pd.DataFrame(buffer)
                    buffer.clear()

            if buffer:
                yield pd.DataFrame(buffer)



import pandas as pd
from typing import Any, Mapping, Optional, List
from parcllabs.common import (
    VALID_LOCATION_TYPES,
    VALID_US_REGIONS,
    VALID_US_STATE_ABBREV,
    VALID_US_STATE_FIPS_CODES,
    VALID_SORT_BY,
    VALID_SORT_ORDER,
    GET_METHOD,
    POST_METHOD,
)
from parcllabs.services.validators import Validators
from parcllabs.services.parcllabs_service import ParclLabsService


class SearchMarkets(ParclLabsService):
    """
    Retrieve parcl_id and metadata for geographic markets in the Parcl Labs API.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _as_pd_dataframe(self, data: List[Mapping[str, Any]]) -> Any:
        return pd.DataFrame(data)

    def retrieve(
        self,
        query: Optional[str] = None,
        location_type: str = None,
        region: str = None,
        state_abbreviation: str = None,
        state_fips_code: str = None,
        parcl_id: int = None,
        geoid: str = None,
        sort_by: str = None,
        sort_order: str = None,
        limit: Optional[int] = None,
        params: Optional[Mapping[str, Any]] = None,
        auto_paginate: bool = False,
    ):
        """
        Retrieve parcl_id and metadata for geographic markets in the Parcl Labs API.

        Args:
            query (str, optional): The search query to filter results by.
            location_type (str, optional): The location type to filter results by.
            region (str, optional): The region to filter results by.
            state_abbreviation (str, optional): The state abbreviation to filter results by.
            state_fips_code (str, optional): The state FIPS code to filter results by.
            parcl_id (int, optional): The parcl_id to filter results by.
            geoid (str, optional): The geoid to filter results by.
            sort_by (str, optional): The field to sort results by.
            sort_order (str, optional): The sort order to apply to the results.
            limit (int, optional): The number of items to return per page.
            params (dict, optional): Additional parameters to include in the request.
            auto_paginate (bool, optional): Automatically paginate through the results.

        Returns:
            Any: The JSON response as a pandas DataFrame.
        """

        params = params or {}

        params = Validators.validate_input_str_param(
            param=location_type,
            param_name="location_type",
            valid_values=VALID_LOCATION_TYPES,
            params_dict=params,
        )

        params = Validators.validate_input_str_param(
            param=region,
            param_name="region",
            valid_values=VALID_US_REGIONS,
            params_dict=params,
        )

        params = Validators.validate_input_str_param(
            param=state_abbreviation,
            param_name="state_abbreviation",
            valid_values=VALID_US_STATE_ABBREV,
            params_dict=params,
        )

        params = Validators.validate_input_str_param(
            param=state_fips_code,
            param_name="state_fips_code",
            valid_values=VALID_US_STATE_FIPS_CODES,
            params_dict=params,
        )

        params = Validators.validate_input_str_param(
            param=sort_by,
            param_name="sort_by",
            valid_values=VALID_SORT_BY,
            params_dict=params,
        )

        params = Validators.validate_input_str_param(
            param=sort_order,
            param_name="sort_order",
            valid_values=VALID_SORT_ORDER,
            params_dict=params,
        )

        if query:
            params["query"] = query

        if parcl_id:
            params["parcl_id"] = parcl_id

        if geoid:
            params["geoid"] = geoid

        if limit:
            params["limit"] = self._validate_limit(GET_METHOD, limit)
        elif self.client.limit:
            params["limit"] = self._validate_limit(GET_METHOD, self.client.limit)

        results = self._fetch_get(
            url=self.full_url, params=params, auto_paginate=auto_paginate
        )
        data = self._as_pd_dataframe(results.get("items"))
        self._update_account_info(results.get("account"))
        self.markets = data
        return data



from collections import deque
from typing import Any, Mapping, Optional, List
from concurrent.futures import ThreadPoolExecutor, as_completed
import pandas as pd


from parcllabs.common import (
    VALID_EVENT_TYPES,
    VALID_ENTITY_NAMES,
)
from parcllabs.enums import RequestLimits
from parcllabs.services.data_utils import (
    safe_concat_and_format_dtypes,
)
from parcllabs.services.validators import Validators
from parcllabs.services.streaming.parcllabs_streaming_service import (
    ParclLabsStreamingService,
)
from parcllabs.exceptions import (
    NotFoundError,
)  # Assuming this is the exception for a 404 error


class PropertyEventsService(ParclLabsStreamingService):
    """
    Retrieve parcl_property_id event history.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retrieve(
        self,
        parcl_property_ids: List[int],
        event_type: str = None,
        start_date: str = None,
        end_date: str = None,
        entity_owner_name: str = None,
        record_updated_date_start: str = None,
        record_updated_date_end: str = None,
        params: Optional[Mapping[str, Any]] = {},
    ):
        """
        Retrieve property events for given parameters.
        """
        params = {}
        params = Validators.validate_input_str_param(
            param=event_type,
            param_name="event_type",
            valid_values=VALID_EVENT_TYPES,
            params_dict=params,
        )

        params = Validators.validate_input_str_param(
            param=entity_owner_name,
            param_name="entity_owner_name",
            valid_values=VALID_ENTITY_NAMES,
            params_dict=params,
        )

        if start_date:
            params["start_date"] = start_date

        if end_date:
            params["end_date"] = end_date

        if record_updated_date_start:
            record_updated_date_start = Validators.validate_date(
                record_updated_date_start
            )
            params["record_updated_date_start"] = record_updated_date_start

        if record_updated_date_end:
            record_updated_date_end = Validators.validate_date(record_updated_date_end)
            params["record_updated_date_end"] = record_updated_date_end

        parcl_property_ids = [str(i) for i in parcl_property_ids]
        total_properties = len(parcl_property_ids)

        def process_batch(batch_ids):
            local_params = params.copy()
            local_params["parcl_property_id"] = batch_ids
            try:
                response = self._post(url=self.full_post_url, data=local_params)
                data = response.json()
                self._update_account_info(data.get("account"))
                return data.get("items")  # Return data as json
            except NotFoundError:
                return None
            except Exception as e:
                print(f"Error processing batch {batch_ids}: {str(e)}")
                return None

        all_data = deque()
        with ThreadPoolExecutor(max_workers=self.client.num_workers) as executor:
            max_post_limit = RequestLimits.MAX_POST.value
            futures = {
                executor.submit(
                    process_batch, parcl_property_ids[i : i + max_post_limit]
                ): len(parcl_property_ids[i : i + max_post_limit])
                for i in range(0, total_properties, max_post_limit)
            }

            for future in as_completed(futures):
                batch_result = future.result()
                if batch_result:
                    batch_df = pd.DataFrame(batch_result)
                    all_data.append(batch_df)

        df = safe_concat_and_format_dtypes(all_data)
        return df



from datetime import datetime
from typing import List
from parcllabs.common import (
    VALID_PORTFOLIO_SIZES,
    VALID_PROPERTY_TYPES,
)


class Validators:
    @staticmethod
    def validate_date(date_str: str) -> str:
        """
        Validates the date string and returns it in the 'YYYY-MM-DD' format.
        Raises ValueError if the date is invalid or not in the expected format.
        """
        if date_str:
            try:
                formatted_date = datetime.strptime(date_str, "%Y-%m-%d").strftime(
                    "%Y-%m-%d"
                )
                return formatted_date
            except ValueError:
                raise ValueError(
                    f"Date {date_str} is not in the correct format YYYY-MM-DD."
                )

    @staticmethod
    def validate_property_type(property_type: str) -> str:
        return Validators._validate_from_list(
            property_type, VALID_PROPERTY_TYPES, "Property type"
        )

    @staticmethod
    def validate_portfolio_size(portfolio_size: str) -> str:
        """
        Validates the portfolio size string and returns it in the expected format.
        Raises ValueError if the portfolio size is invalid or not in the expected format.
        """
        return Validators._validate_from_list(
            portfolio_size, VALID_PORTFOLIO_SIZES, "Portfolio size"
        )

    @staticmethod
    def _validate_from_list(value: str, valid_list: List[str], value_type: str) -> str:
        if value:
            value = value.strip().upper()
            if value.lower() not in [v.lower() for v in valid_list]:
                raise ValueError(
                    f"{value_type} {value} is not valid. Must be one of {', '.join(valid_list)}."
                )
        return value

    @staticmethod
    def validate_input_str_param(
        param: str, param_name: str, valid_values: List[str], params_dict: dict = None
    ):
        if param:

            param = param.upper()

            params_dict[param_name] = param

            if param not in valid_values:
                raise ValueError(
                    f"{param_name} value error. Valid values are: {valid_values}. Received: {param}"
                )

        return params_dict

    @staticmethod
    def validate_input_bool_param(param, param_name: str, params_dict: dict = None):
        if param is None:
            return params_dict  # or return None if that’s preferred

        if not isinstance(param, bool):
            raise ValueError(
                f"{param_name} value error. Expected boolean. Received: {param}"
            )

        params_dict[param_name] = "true" if param else "false"
        return params_dict

    @staticmethod
    def validate_parcl_ids(parcl_ids):
        if isinstance(parcl_ids, int):
            parcl_ids = [parcl_ids]
        return parcl_ids

    @staticmethod
    def validate_us_zip_code(zip_code: str) -> str:
        """
        Validates the US zip code string and returns it in the expected format.
        Raises ValueError if the zip code is invalid or not in the expected format.
        """
        zip_code = zip_code.strip()
        if not zip_code.isdigit() or len(zip_code) != 5:
            raise ValueError(f"Zip code {zip_code} is not a valid 5-digit US zip code.")
        return zip_code

    @staticmethod
    def validate_field_exists(data: dict, fields: List[str]):
        """
        Validates that the required fields exist in the provided dictionary.
        """
        missing_fields = [field for field in fields if field not in data.keys()]
        if missing_fields:
            raise ValueError(
                f"Missing required fields: {', '.join(missing_fields)}. Provided request: {data}"
            )



from typing import List
from collections import deque

import pandas as pd

from parcllabs.common import VALID_PROPERTY_TYPES_UNIT_SEARCH, VALID_ENTITY_NAMES
from parcllabs.services.validators import Validators
from parcllabs.services.streaming.parcllabs_streaming_service import (
    ParclLabsStreamingService,
)


class PropertySearch(ParclLabsStreamingService):
    """
    Retrieve parcl_property_id for geographic markets in the Parcl Labs API.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retrieve(
        self,
        parcl_ids: List[int],
        property_type: str,
        square_footage_min: int = None,
        square_footage_max: int = None,
        bedrooms_min: int = None,
        bedrooms_max: int = None,
        bathrooms_min: int = None,
        bathrooms_max: int = None,
        year_built_min: int = None,
        year_built_max: int = None,
        current_entity_owner_name: str = None,
        event_history_sale_flag: bool = None,
        event_history_rental_flag: bool = None,
        event_history_listing_flag: bool = None,
        current_new_construction_flag: bool = None,
        current_owner_occupied_flag: bool = None,
        current_investor_owned_flag: bool = None,
        record_added_date_start: str = None,
        record_added_date_end: str = None,
    ):
        params = {}

        params = Validators.validate_input_str_param(
            param=property_type,
            param_name="property_type",
            valid_values=VALID_PROPERTY_TYPES_UNIT_SEARCH,
            params_dict=params,
        )

        params = Validators.validate_input_str_param(
            param=current_entity_owner_name,
            param_name="current_entity_owner_name",
            valid_values=VALID_ENTITY_NAMES,
            params_dict=params,
        )

        params = Validators.validate_input_bool_param(
            param=event_history_sale_flag,
            param_name="event_history_sale_flag",
            params_dict=params,
        )

        params = Validators.validate_input_bool_param(
            param=event_history_rental_flag,
            param_name="event_history_rental_flag",
            params_dict=params,
        )

        params = Validators.validate_input_bool_param(
            param=event_history_listing_flag,
            param_name="event_history_listing_flag",
            params_dict=params,
        )

        params = Validators.validate_input_bool_param(
            param=current_new_construction_flag,
            param_name="current_new_construction_flag",
            params_dict=params,
        )

        params = Validators.validate_input_bool_param(
            param=current_owner_occupied_flag,
            param_name="current_owner_occupied_flag",
            params_dict=params,
        )

        params = Validators.validate_input_bool_param(
            param=current_investor_owned_flag,
            param_name="current_investor_owned_flag",
            params_dict=params,
        )

        if bedrooms_max:
            params["bedrooms_max"] = bedrooms_max

        if bedrooms_min:
            params["bedrooms_min"] = bedrooms_min

        if bathrooms_max:
            params["bathrooms_max"] = bathrooms_max

        if bathrooms_min:
            params["bathrooms_min"] = bathrooms_min

        if year_built_max:
            params["year_built_max"] = year_built_max

        if year_built_min:
            params["year_built_min"] = year_built_min

        if square_footage_max:
            params["square_footage_max"] = square_footage_max

        if square_footage_min:
            params["square_footage_min"] = square_footage_min

        if record_added_date_start:
            record_added_date_start = Validators.validate_date(record_added_date_start)
            params["record_added_date_start"] = record_added_date_start

        if record_added_date_end:
            record_added_date_end = Validators.validate_date(record_added_date_end)
            params["record_added_date_end"] = record_added_date_end

        output_data = deque()
        total_parcl_ids = len(parcl_ids)

        for parcl_id in parcl_ids:
            params["parcl_id"] = parcl_id
            response = self._get(url=self.full_url, params=params)
            data = response.json()
            df_container = pd.DataFrame(data.get("items"))
            self._update_account_info(data.get("account"))
            output_data.append(df_container)
        results = pd.concat(output_data).reset_index(drop=True)
        return results



from typing import Optional
from parcllabs.__version__ import VERSION as __version__

# Constants
DEFAULT_API_BASE = "https://api.parcllabs.com"

api_key: Optional[str] = None
api_base = DEFAULT_API_BASE


from parcllabs.parcllabs_client import ParclLabsClient
